from __future__ import annotations

import abc
from pathlib import Path
from typing import Any, overload

import numpy as np
import polars as pl
from polars import ColumnNotFoundError

from neuralib.util.verbose import fprint, printdf
from .baselog import STIMPY_SOURCE_VERSION

__all__ = ['AbstractStimProtocol']


class AbstractStimProtocol(metaclass=abc.ABCMeta):
    """ABC class for the stimpy protocol file (.prot)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = Number of Stim Type

        C = Numbers of Cycle
    """

    name: str
    """protocol name. related to filename"""

    options: dict[str, Any]
    """protocol options"""

    visual_stimuli_dataframe: pl.DataFrame
    """visual stimuli dataframe. row number: S"""

    version: STIMPY_SOURCE_VERSION
    """date of major changes"""

    def __repr__(self):
        ret = list()

        ret.append('# general parameters')
        for k, v in self.options.items():
            ret.append(f'{k} = {v}')
        ret.append('# stimulus conditions')
        ret.append('\t'.join(self.stim_params))

        ret.append(printdf(self.visual_stimuli_dataframe, do_print=False))

        return '\n'.join(ret)

    def __init__(self, name: str,
                 options: dict[str, Any],
                 visual_stimuli: pl.DataFrame,
                 version: STIMPY_SOURCE_VERSION):

        self.name = name
        self.options = options
        self.visual_stimuli_dataframe = visual_stimuli
        self.version = version

        self._visual_stimuli = None

    @property
    def n_stimuli(self) -> int:
        """number of stimuli (S)"""
        return self.visual_stimuli_dataframe.shape[0]

    @property
    def stim_params(self) -> tuple[str, ...]:
        return tuple(self.visual_stimuli_dataframe.columns)

    @overload
    def __getitem__(self, item: int) -> pl.DataFrame:
        """get row of visual stimuli"""
        pass

    @overload
    def __getitem__(self, item: str) -> np.ndarray:
        """get header of stimuli"""
        pass

    def __getitem__(self, item: str | int) -> np.ndarray | pl.DataFrame:
        """Get protocol value of parameter *item*

        :param item: parameter name. either the row number(int) or field name (str)
        :return: protocol value
        :raises TypeError: `item` not a `str` or `int`
        """
        if isinstance(item, str):
            try:
                return self.visual_stimuli_dataframe.get_column(item).to_numpy()
            except ColumnNotFoundError as e:
                fprint(f'INVALID: {item}, select from {tuple(self.visual_stimuli_dataframe.columns)}', vtype='error')
                raise e

        elif isinstance(item, int):
            ret = {
                h: self.visual_stimuli_dataframe.row(item)[i]
                for i, h in enumerate(self.visual_stimuli_dataframe.columns)
            }
            return pl.DataFrame(ret)
        else:
            raise TypeError(f'{type(item)}')

    @classmethod
    @abc.abstractmethod
    def load(cls, file: Path | str) -> AbstractStimProtocol:
        r"""
        Load \*.prot file

        :param file: file path
        """
        pass

    @property
    @abc.abstractmethod
    def is_shuffle(self) -> bool:
        """if shuffle stimulation"""
        pass

    @property
    @abc.abstractmethod
    def background(self) -> float:
        """background for non-stimulation epoch.
        Note the default value need to check in user-specific stimpy version
        """
        pass

    @property
    @abc.abstractmethod
    def start_blank_duration(self) -> int:
        """blank duration before starting the visual stimulation epoch (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def blank_duration(self) -> int:
        """blank duration between each visual stimulus (in sec)"""
        pass

    @property
    @abc.abstractmethod
    def trial_blank_duration(self) -> int:
        """blank duration between trials (in sec)
        TODO check stimpy source code"""
        pass

    @property
    @abc.abstractmethod
    def end_blank_duration(self) -> int:
        """blank duration after the visual stimulation epoch"""
        pass

    @property
    @abc.abstractmethod
    def trial_duration(self) -> int:
        """trial duration"""
        pass

    @property
    def visual_duration(self) -> int:
        """total visual duration"""
        return self.trial_duration * self.n_trials

    @property
    def total_duration(self) -> int:
        """total protocol duration"""
        return self.start_blank_duration + self.visual_duration + self.end_blank_duration

    @property
    def stimulus_type(self) -> str:
        """stimulus type"""
        return self.options['stimulusType']

    @property
    def n_trials(self) -> int:
        """(T,)"""
        return self.options['nTrials']




from __future__ import annotations

import abc
from pathlib import Path
from typing import Literal, TypeVar, Generic, cast, TypedDict, Any

import numpy as np
import polars as pl

from neuralib.typing import PathLike
from .event import RigEvent, CamEvent
from .preference import PreferenceDict, load_preferences
from .session import Session, SessionInfo
from .stimulus import AbstractStimulusPattern

__all__ = [
    'STIMPY_SOURCE_VERSION',
    'LOG_SUFFIX',
    'CAMERA_TYPE',
    #
    'Baselog',
    'StimlogBase',
]

STIMPY_SOURCE_VERSION = Literal['pyvstim', 'stimpy-bit', 'stimpy-git', 'debug']
LOG_SUFFIX = Literal['.log', '.riglog']
CAMERA_TYPE = Literal['facecam', 'eyecam', '1P_cam']

#
R = TypeVar('R', bound='Baselog')  # Riglog-Like
S = TypeVar('S', bound='StimlogBase')  # Stimlog-Like
P = TypeVar('P', bound='AbstractStimProtocol')  # protocol file


# =========== #
# Baselog ABC #
# =========== #

class RigConfig(TypedDict, total=False):
    source_version: STIMPY_SOURCE_VERSION
    """stimpy source version {'pyvstim', 'stimpy-bit', 'stimpy-git'}"""
    version: float
    """acquisition flag. i.e., 0.3"""
    commit_hash: str
    """git commit hash"""
    codes: dict[str, int]
    """<EVENT_TYPES>:<NUMBER>"""
    fields: tuple[str, ...]
    """column repr for the logging"""


class Baselog(Generic[S, P], metaclass=abc.ABCMeta):
    """ABC class for different stimpy/pyvstim log files. i.e., .log, .riglog"""

    log_config: RigConfig

    def __init__(self,
                 root_path: PathLike,
                 log_suffix: LOG_SUFFIX,
                 diode_offset: bool = True,
                 reset_mapping: dict[int, list[str]] | None = None):
        """
        :param root_path: log file path or log directory
        :param log_suffix: log file suffix
        :param diode_offset: whether do the diode offset
        """

        if not isinstance(root_path, Path):
            root_path = Path(root_path)

        if root_path.is_dir():
            self.riglog_file = self._find_logfile(root_path, log_suffix)
        else:
            self.riglog_file = root_path

        self.log_config = self._get_log_config()
        self.version = self.log_config['source_version']

        self.dat = self._cache_asarray(self.riglog_file)

        #
        self._diode_offset = diode_offset
        self._reset_mapping = reset_mapping

    @classmethod
    def _find_logfile(cls,
                      root: Path,
                      log_suffix: LOG_SUFFIX) -> Path:

        f = list(root.glob(f'*{log_suffix}'))
        if len(f) == 1:
            return f[0]

        elif len(f) == 0:
            print(f'no riglog under {root}, try to find in the subfolder...')
            for s in root.iterdir():
                if s.is_dir() and s.name.startswith('run0'):
                    try:
                        return cls._find_logfile(s, log_suffix)
                    except FileNotFoundError:
                        pass

            raise FileNotFoundError(f'no riglog file {log_suffix} under {root}')

        else:
            raise FileNotFoundError(f'more than one riglog files under {root}')

    # noinspection PyTypedDict
    def _get_log_config(self) -> RigConfig:
        """get config dict and source version for different logs from the # headers"""
        ret = RigConfig()
        with open(self.riglog_file) as f:
            for line in f:
                if '#' in line:
                    if 'RIG VERSION' in line:
                        ret['version'] = float(line.split(': ')[-1])

                    elif 'RIG GIT COMMIT HASH' in line:
                        ret['commit_hash'] = line.split(': ')[-1].strip()

                    elif 'CODES' in line:
                        codes = {}
                        content = line.replace('# CODES: ', '').strip()
                        iter_codes = content.split(',')
                        for pair in iter_codes:
                            code, num = pair.split('=')
                            code = code.strip()
                            value = int(num.strip())
                            codes[code.lower()] = value

                        ret['codes'] = codes

                    elif 'RIG CSV' in line:
                        content = line.replace('# RIG CSV: ', '').strip()
                        ret['fields'] = tuple(content.split(','))

        # infer
        if 'opto' not in ret['codes']:
            ret['source_version'] = 'stimpy-bit'

        if 'version' not in ret:
            ret['source_version'] = 'stimpy-git'

        if 'opto' in ret['codes'] and ret['codes']['opto'] == 15:
            ret['source_version'] = 'pyvstim'

        return ret

    @classmethod
    @abc.abstractmethod
    def _cache_asarray(cls, filepath: Path) -> np.ndarray:
        pass

    # ============ #
    # RIGLOG EVENT #
    # ============ #

    def _event(self, code: int) -> np.ndarray:
        """shape(sample, 2) with time and value"""
        x = self.dat[:, 0] == code
        t = self.dat[x, 2].copy() / 1000  # s
        v = self.dat[x, 3].copy()
        return np.vstack([t, v]).T

    @property
    def exp_start_time(self) -> float:
        """experimental start time (in sec)"""
        return float(self.dat[0, 2].copy() / 1000)

    @property
    def exp_end_time(self) -> float:
        """experimental end time (in sec)"""
        return float(self.dat[-1, 2].copy() / 1000)

    @property
    def total_duration(self) -> float:
        """experimental duration (in sec)"""
        return self.exp_end_time - self.exp_start_time

    @property
    def screen_event(self) -> RigEvent:
        """screen rig event. i.e., diode pulse"""
        return RigEvent('screen', self._event(0))

    @property
    def imaging_event(self) -> RigEvent:
        """imaging rig event. i.e., 2photon pulse"""
        return RigEvent('imaging', self._event(1))

    @property
    def position_event(self) -> RigEvent:
        """position rig event. i.e., encoder pulse"""
        return RigEvent('position', self._event(2))

    @property
    def lick_event(self) -> RigEvent:
        """lick rig event. i.e., lick meter pulse"""
        return RigEvent('lick', self._event(3))

    @property
    def reward_event(self) -> RigEvent:
        """reward rig event. i.e., reward given pulse from lick meter"""
        if self.version == 'stimpy-git':
            return RigEvent('reward', self._event(5))
        else:
            return RigEvent('reward', self._event(4))

    @property
    def lap_event(self) -> RigEvent:
        """lap rig event. i.e., optic sensing for the reflective taps"""
        if self.version == 'stimpy-git':
            return RigEvent('lap', self._event(6))
        else:
            return RigEvent('lap', self._event(5))

    @property
    def act_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    @property
    def opto_event(self) -> RigEvent:
        """todo"""
        raise NotImplementedError('')

    class CameraEvent:
        """camera event"""
        camera: dict[CAMERA_TYPE, int]

        def __init__(self, rig: R):
            """
            :param rig:``Baselog``
            """
            self.rig = rig

            if rig.version == 'stimpy-git':
                self.camera = {
                    'facecam': 7,
                    'eyecam': 8,
                    '1P_cam': 9,
                }
            else:
                self.camera = {
                    'facecam': 6,
                    'eyecam': 7,
                    '1P_cam': 8,
                }

        def __getitem__(self, item: CAMERA_TYPE) -> CamEvent:
            if item not in self.camera:
                raise KeyError('cam id not found')
            return CamEvent(item, self.rig._event(self.camera[item]))

    @property
    def camera_event(self) -> CameraEvent:
        """camera event. including {'facecam', 'eyecam', '1P_cam'} implemented by __getitem__()"""
        return self.CameraEvent(self)

    # ========================= #
    # Access Other output Files #
    # ========================= #

    @property
    def prot_file(self) -> Path:
        """protocol file path"""
        return self.riglog_file.with_suffix('.prot')

    @abc.abstractmethod
    def get_protocol(self) -> P:
        """
        get protocol (TypeVar ``P``)

        :return: :class:`~stimpyp.parser.baseprot.AbstractStimProtocol()`
        """
        pass

    def get_stimulus_type(self) -> str:
        return self.get_protocol().stimulus_type

    @property
    def pref_file(self) -> Path:
        """preferences file path"""
        return self.riglog_file.with_suffix('.prefs')

    def get_preferences(self) -> PreferenceDict:
        """get preferences file"""
        return load_preferences(self.pref_file)

    @abc.abstractmethod
    def get_stimlog(self) -> S:
        """get stimlog (TypeVar ``S``)

        :return: :class:`~stimpyp.parser.baselog.StimlogBase()`
        """
        pass


# =========== #
# Stimlog ABC #
# =========== #

class StimlogBase(Generic[R], metaclass=abc.ABCMeta):
    """ABC for stimulation logging. i.e., .log or .stimlog

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        P = number of acquisition sample pulse

    """

    # =========== #
    # Log Headers #
    # =========== #

    config: dict[str, Any] = {}
    """i.e., name, commit hash, missed_frames, ..."""

    log_info: dict[int, str] = {}
    """i.e., {10: 'vstim', 20: 'stateMachine'}"""

    log_header: dict[int, list[str]] = {}
    """i.e., {10: ['code','presentTime','iStim', ...], 20: ['code', 'elapsed', 'cycle', ...]}"""

    # =========== #
    # Common Attr #
    # =========== #

    time: np.ndarray
    """acquisition time in sec. Array[float, P]"""

    stim_index: np.ndarray
    """stimulation index. Array[int, P]"""

    trial_index: np.ndarray
    """trial index. Array[int, P]"""

    contrast: np.ndarray
    """stimulus contrast. Array[int, P]. value domain in (0,1)"""

    frame_index: np.ndarray
    """stimulation frame index, recount every N. Array[int, P]"""

    # ======= #
    # Grating #
    # ======= #

    ori: np.ndarray
    """direction in deg. Array[float, P]"""

    sf: np.ndarray
    """spatial frequency in cyc/deg. Array[float, P]"""

    tf: np.ndarray
    """temporal frequency in hz. Array[float, P]"""

    phase: np.ndarray
    """stimulus phase for each N. Array[float, P]"""

    # ============= #
    # Function Base #
    # ============= #

    pos_x: np.ndarray
    """object center position X. Array[float, P]"""

    pos_y: np.ndarray
    """object center position Y. Array[float, P]"""

    size_x: np.ndarray
    """object size width. Array[int, P]"""

    size_y: np.ndarray
    """object size height. Array[int, P]"""

    # ====== #
    # Others #
    # ====== #

    flick: np.ndarray
    """TODO. Array[int, P]"""

    interpolate: np.ndarray
    """whether do the interpolate, Array[bool, P]"""

    mask: np.ndarray
    """TODO. Array[bool|None, P]"""

    pattern: np.ndarray
    """object pattern. Array[str, P]"""

    def __init__(self,
                 riglog: R,
                 file_path: PathLike | None,
                 reset_mapping: dict[int, list[str]] | None = None):
        """
        :param riglog: :class:`Baselog`
        :param file_path: filepath of stimlog. could be None if shared log (pyvstim case)
        :param reset_mapping: Customized mapping for ``reset()``
            - key: corresponding to :attr:`log_header`
            - value: list of field, should be the same name as class annotations
        """
        self.riglog_data = riglog
        if file_path is not None:
            self.stimlog_file = Path(file_path)

        self._reset_mapping = reset_mapping

    @abc.abstractmethod
    def _reset(self) -> None:
        """used for assign attributes"""
        pass

    def _reset_cust_mapping(self):
        pass

    # ============ #
    # As Dataframe #
    # ============ #

    @abc.abstractmethod
    def get_visual_stim_dataframe(self, **kwargs) -> pl.DataFrame:
        pass

    @abc.abstractmethod
    def get_state_machine_dataframe(self) -> pl.DataFrame:
        pass

    # ========= #
    # Time Info #
    # ========= #

    @property
    @abc.abstractmethod
    def exp_start_time(self) -> float:
        """experimental start time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    @abc.abstractmethod
    def exp_end_time(self) -> float:
        """experimental end time (in sec, synced to riglog time with diode offset)"""
        pass

    @property
    def stim_start_time(self) -> float:
        """the first stimulation start time (in sec, synced to riglog time with diode offset)"""
        return cast(float, self.stimulus_segment[0, 0])

    @property
    def stim_end_time(self) -> float:
        """the last stimulation end time (in sec, synced to riglog time with diode offset)"""
        return cast(float, self.stimulus_segment[-1, 1])

    @property
    @abc.abstractmethod
    def stimulus_segment(self) -> np.ndarray:
        """stimulation time segment (on-off) in sec (N, 2)"""
        pass

    def stim_square_pulse_event(self, sampling_rate: float = 30.) -> RigEvent:
        """
        Get the stimulation on-off square pulse 0,1 consecutive event

        :param sampling_rate: sampling rate for the time domain interpolation
        :return: Stimulus rig event
        """
        start_time = self.exp_start_time
        end_time = self.exp_end_time
        seg = self.stimulus_segment

        t = np.arange(start_time, end_time, 1 / sampling_rate)
        ret = np.zeros_like(t)
        for (on, off) in seg:
            mask = np.logical_and(on < t, t < off)
            ret[mask] = 1

        return RigEvent('visual_stim', np.vstack((t, ret)).T)

    @abc.abstractmethod
    def session_trials(self) -> dict[Session, SessionInfo]:
        """get the session:SessionInfo dictionary (experimental and user-specific)"""
        pass

    @property
    @abc.abstractmethod
    def time_offset(self) -> float:
        """time (in sec) to sync stimlog time to riglog"""
        pass

    # =========================== #
    # Stim Trial/Index/Cycle Info #
    # =========================== #

    @property
    def n_cycles(self) -> list[int]:
        """Number of cycle for each trial"""
        raise [1]

    @property
    @abc.abstractmethod
    def profile_dataframe(self) -> pl.DataFrame:
        """
        Dataframe with columns:

        - stim type index: ``i_stims``
        - trial index: ``i_trials``
        """
        pass

    # ================= #
    # Stim Pattern Info #
    # ================= #

    @abc.abstractmethod
    def get_stim_pattern(self, **kwargs) -> AbstractStimulusPattern:
        """get pattern foreach stimulation"""
        pass



"""
Stimpy parser
==============

:author:
    Yu-Ting Wei


This module provide parser utilities for data acquired from stimpy (lab internal usage)

"""

from stimpyp.parser.baselog import *
from stimpyp.parser.baseprot import *
from stimpyp.parser.camlog import *
from stimpyp.parser.event import *
from stimpyp.parser.session import *
from stimpyp.parser.stimpy_core import *
from stimpyp.parser.stimpy_git import *
from stimpyp.parser.stimpy_pyv import *
from stimpyp.parser.stimulus import *
from stimpyp.parser.util import *



from __future__ import annotations

import abc
from typing import NamedTuple, Iterable, TYPE_CHECKING, TypeVar, Any, Generic

import numpy as np
from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from stimpyp.parser.baselog import R

__all__ = [
    'Direction',
    'SF',
    'TF',
    'SFTF',
    'VisualParas',
    #
    'GratingStim',
    'FunctionStim',
    #
    'AbstractStimulusPattern',
    'GratingPattern',
    'FunctionPattern'
]

Direction: TypeAlias = int
"""stimulus direction in deg"""

SF: TypeAlias = float
"""spatial frequency in cyc/deg"""

TF: TypeAlias = int
"""temporal frequency in hz"""

SFTF: TypeAlias = tuple[SF, TF]
"""tuple of spatial and temporal frequencies"""

VisualParas: TypeAlias = tuple[SF, TF, Direction]
"""tuple of spatial, temporal frequencies and degree"""


# =============== #
# Individual Stim #
# =============== #


class GratingStim(NamedTuple):
    index: int
    """stimulus index"""
    time: np.ndarray
    """stim on-off time. Array[float, 2]."""
    sf: SF
    """``SF``"""
    tf: TF
    """``TF``"""
    direction: Direction
    """``Direction``"""


class FunctionStim(NamedTuple):
    index: int
    """stimulus index"""
    time: np.ndarray
    """stim on-off time. Array[float, 2]."""
    pos_xy: np.ndarray
    """object center position XY. Array[float, 2]"""
    size_xy: np.ndarray
    """object size width and height. Array[float, 2]"""


# ================ #
# Stimulus Pattern #
# ================ #

S = TypeVar('S')  # Individual Stim
P = TypeVar('P')  # pattern


class AbstractStimulusPattern(Generic[P, S], metaclass=abc.ABCMeta):
    """
    Abstract Stimulus Pattern

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)
    """

    time: np.ndarray
    """stim on-off in sec. Array[float, [N, 2]]"""
    contrast: np.ndarray
    """stimulus contrast. Array[float, N]"""
    duration: np.ndarray
    """theoretical duration in prot file, not actual detected using diode. Array[float, N]"""

    def __init__(self,
                 time: np.ndarray,
                 contrast: np.ndarray,
                 *,
                 duration: np.ndarray | None = None):
        """

        :param time: stim on-off in sec. Array[float, [N, 2]]
        :param contrast: stimulus contrast. Array[float, N]
        :param duration: theoretical duration in prot file, not actual detected using diode. Array[float, N]
        """
        self.time = time
        self.contrast = contrast
        self.duration = duration

    @classmethod
    def of(cls, rig: 'R') -> P:
        """
        init from Baselog children class

        :param rig: :class:`~stimpyp.parser.baselog.Baselog`
        :return: :class:`StimPattern`
        """
        return rig.get_stimlog().get_stim_pattern()

    @abc.abstractmethod
    def foreach_stimulus(self, name: bool = False) -> Iterable[Any | S]:
        pass


class GratingPattern(AbstractStimulusPattern):
    """Grating Stimulus Pattern"""

    direction: np.ndarray
    """stimulus direction in deg. Array[int, N]"""

    sf: np.ndarray
    """spatial frequency in cyc/deg. Array[float, N]"""

    tf: np.ndarray
    """temporal frequency in hz Array[int, N]"""

    def __init__(self,
                 time: np.ndarray,
                 contrast: np.ndarray,
                 direction: np.ndarray,
                 sf: np.ndarray,
                 tf: np.ndarray,
                 *,
                 duration: np.ndarray | None = None):
        """

        :param time: stim on-off in sec. Array[float, [N, 2]]
        :param contrast: stimulus contrast. Array[float, N]
        :param direction: stimulus direction in deg. Array[int, N]
        :param sf: spatial frequency in cyc/deg. Array[float, N]
        :param tf: temporal frequency in hz Array[int, N]
        :param duration: theoretical duration in prot file, not actual detected using diode. Array[float, N]
        """

        super().__init__(time, contrast, duration=duration)

        self.direction = direction
        self.sf = sf
        self.tf = tf

    @property
    def sf_set(self) -> np.ndarray:
        """unique sf_set"""
        return np.array(sorted(self.sf_i().keys()))

    @property
    def tf_set(self) -> np.ndarray:
        """unique tf_set"""
        return np.array(sorted(self.tf_i().keys()))

    @property
    def n_sf(self) -> int:
        """number of sf set"""
        return len(self.sf_set)

    @property
    def n_tf(self) -> int:
        """number of tf set"""
        return len(self.tf_set)

    @property
    def n_sftf(self) -> int:
        """number of sftf combination"""
        return len(self.sftf_i())

    @property
    def n_dir(self) -> int:
        """number of direction"""
        return len(self.dir_i())

    def dir_i(self) -> dict[Direction, int]:
        """deg:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.direction)))}

    def sf_i(self) -> dict[SF, int]:
        """sf:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.sf)))}

    def tf_i(self) -> dict[TF, int]:
        """sf:index dict"""
        return {it: i for i, it in enumerate(sorted(np.unique(self.tf)))}

    # previous plot use tfsf as condition idx
    def sftf_i(self) -> dict[SFTF, int]:
        """sf, tf combination. (sf , tf):y"""
        return {
            it: i
            for i, it in enumerate([
                (sf, tf)
                for sf in sorted(np.unique(self.sf))
                for tf in sorted(np.unique(self.tf))
            ])
        }

    def sftfdir_i(self) -> dict[VisualParas, int]:
        """sf, tf, ori combination. (sf , tf , ori):y"""
        return {
            it: i  # (sf , tf, ori): index
            for i, it in enumerate([
                (sf, tf, ori * 30)
                for sf in sorted(np.unique(self.sf))
                for tf in sorted(np.unique(self.tf))
                for ori in range(12)
            ])
        }

    def foreach_stimulus(self, name: bool = False) -> Iterable[tuple[int, np.ndarray, SF, TF, Direction] | GratingStim]:
        """
        Generator for (index, stimulus_time, sf, tf, ori)

        :param name: If True, return ``GratingStim``, otherwise, return tuple
        :return:
        """
        for si, st in enumerate(self.time):
            ret = si, st, self.sf[si], self.tf[si], self.direction[si]

            if name:
                yield GratingStim(*ret)
            else:
                yield ret

    def get_stim_time(self) -> float:
        """get approximate stim time if the same duration. i.e., for plotting purpose"""
        return np.mean(self.time[:, 1] - self.time[:, 0])


class FunctionPattern(AbstractStimulusPattern):
    """"Function Stimulus Pattern"""

    pos_xy: np.ndarray
    """object center position XY. Array[float, [N, 2]]"""
    size_xy: np.ndarray
    """object size width and height. Array[float, [N, 2]]"""

    def __init__(self, time: np.ndarray,
                 contrast: np.ndarray,
                 pos_xy: np.ndarray,
                 size_xy: np.ndarray, *,
                 duration: np.ndarray | None = None):
        """

        :param time: stim on-off in sec. Array[float, [N, 2]]
        :param contrast: stimulus contrast. Array[float, N]
        :param pos_xy: object center position XY. Array[float, [N, 2]]
        :param size_xy: object size width and height. Array[float, [N, 2]]
        :param duration: theoretical duration in prot file, not actual detected using diode. Array[float, N]
        """
        super().__init__(time, contrast, duration=duration)

        self.pos_xy = pos_xy
        self.size_xy = size_xy

    def foreach_stimulus(
            self,
            name: bool = False
    ) -> Iterable[tuple[int, np.ndarray, np.ndarray, np.ndarray] | FunctionStim]:
        """
        Generator for (index, stimulus_time, pos_xy, size_xy)

        :param name: If True, return ``FunctionStim``, otherwise, return tuple
        :return:
        """

        for si, st in enumerate(self.time):
            ret = si, st, self.pos_xy[si], self.size_xy[si]

            if name:
                yield FunctionStim(*ret)
            else:
                yield ret



from __future__ import annotations

import re
from pathlib import Path
from typing import Any, Callable, final

import numpy as np
import polars as pl

from neuralib.typing import PathLike
from neuralib.util.deprecation import deprecated_func
from neuralib.util.verbose import fprint
from .baselog import StimlogBase
from .session import Session, SessionInfo, get_protocol_sessions
from .stimpy_core import RiglogData, StimpyProtocol
from .stimulus import GratingPattern

__all__ = ['StimlogGit',
           'load_stimlog']


@final
class StimlogGit(StimlogBase):
    """class for handle the stimlog file for stimpy **github** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        I = number of photo indicator pulse
    """

    log_info: dict[int, str]
    """{0: <'Gratings', ...>, 1: 'PhotoIndicator', 2: 'StateMachine', 3: 'LogDict'}"""
    log_header: dict[int, list[str]]
    """list of header"""
    log_data: dict[int, list[tuple[Any, ...]]]
    """only for StateMachine(2) and LogDict(3)"""

    # =========== #
    # Visual Pars #
    # =========== #

    duration: np.ndarray
    """duration in sec. Array[float, P]"""
    pos_xy: np.ndarray
    """object center position XY. Array[float, [P, 2]]"""
    size_xy: np.ndarray
    """object size width and height. Array[int, [P, 2]]"""

    # ============== #
    # PhotoIndicator #
    # ============== #

    photo_time: np.ndarray
    """photoindicator time. Array[float, I]"""
    photo_state: np.ndarray
    """photoindicator state. Array[bool, I]"""
    photo_size: np.ndarray
    """photoindicator size in given unit. Array[float, I]"""
    photo_pos: np.ndarray
    """photoindicator in XY. Array[int, [I, 2]]"""
    photo_units: np.ndarray
    """photoindicator unit. Array[str, I]"""
    photo_mode: np.ndarray
    """photoindicator mode. Array[int, I]"""
    photo_frames: np.ndarray
    """photoindicator frames . Array[int, I]"""
    photo_enable: np.ndarray
    """photoindicator size in pixel. Array[bool, I]"""

    def __init__(self, riglog: RiglogData,
                 file_path: PathLike,
                 reset_mapping: dict[int, list[str]] | None = None,
                 diode_offset: bool = True):

        super().__init__(riglog, file_path, reset_mapping)
        self.diode_offset = diode_offset
        self._cache_time_offset: float | None = None

        self._reset()

    def _reset(self):
        log_data = {}

        with self.stimlog_file.open() as f:
            for line, content in enumerate(f):  # type: int, str
                content = content.strip()
                # `## (CODE VERSION) : (commit hash: 88c4705 - tags: [''])`
                # *: 0-inf; +: 1-inf; ?: 0-1
                m = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
                if m:
                    info_name = m.group(1)
                    info_value = m.group(2)

                    if info_name == 'LOG NAME':
                        self.config['log_name'] = info_value
                    elif info_name == 'CODE VERSION':
                        self._reset_code_version(info_value)
                    elif info_name == 'Format':
                        self.config['format'] = info_value.split(' ')
                        if self.config['format'] != ['source_id', 'time', 'source_infos']:
                            raise RuntimeError('stimlog format changed')
                    elif info_name == 'Rig trigger on':
                        info_value = info_value.split(',')
                        self.config['rig_trigger'] = (info_value[0], float(info_value[1]))
                    elif self._reset_log_info(info_name, info_value):
                        pass
                    else:
                        print(f'ignore header line at {line + 1} : {content}')

                elif content.startswith('### START'):
                    self.config['start_time'] = float(content[9:].strip())
                elif content.startswith('### END'):
                    self.config['end_time'] = float(content[7:].strip())
                elif content.startswith('# Missed') and content.endswith('frames'):
                    self.config['missed_frames'] = int(content.split(' ')[2])
                elif content.startswith('#'):
                    print(f'ignore header line at {line + 1} : {content}')

                else:
                    try:
                        self._reset_line(log_data, line + 1, content)
                    except BaseException:
                        raise RuntimeError(f'bad format line at {line + 1} : {content}')

        self.log_data = self._reset_final(log_data)

    def _reset_code_version(self, info_value: str):
        """for ### CODE VERSION,  commit hash and tags"""
        info_value = info_value.strip().split(' ')
        commits = info_value[info_value.index('hash:') + 1]
        tag = info_value[info_value.index('tags:') + 1]

        self.config['commit_hash'] = commits if commits != 'None' else None
        self.config['tag'] = tag if tag != 'None' else None

    def _reset_log_info(self, code: str, info_value: str):
        try:
            code = int(code)
        except ValueError:
            return False

        name, header = info_value.split(' ', maxsplit=1)
        header = eval(header)
        self.log_info[code] = name
        self.log_header[code] = header
        return True

    def _reset_line(self, log_data: dict[int, list], line: int, content: str):
        message: str
        value: list
        code, time, message = content.split(' ', maxsplit=2)
        code = int(code)
        time = float(time)

        if self.log_info[code] in ('Gratings', 'FunctionBased', 'PhotoIndicator', 'LogDict'):
            value = eval(message)
            if len(self.log_header[code]) != len(value):
                print(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        elif self.log_info[code] == 'StateMachine':
            value = eval(message.replace('<', '("').replace(':', '",').replace('>', ')'))
            value = list(map(str, value))
            if len(self.log_header[code]) != len(value):
                print(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        else:
            print(f'unknown log category : at line {line} : {content}')

    def _reset_final(self, log_data: dict[int, list]) -> dict[int, list]:
        remove_code = []
        for code, content in log_data.items():
            if self.log_info[code] in ('Gratings', 'FunctionBased'):
                self.time = np.array([it[0] for it in content])
                self.duration = np.array([it[1] for it in content])
                self.contrast = np.array([it[2] for it in content])
                self.ori = np.array([it[3] for it in content])
                self.phase = np.array([it[4] for it in content])
                self.pos_xy = np.array([it[5] for it in content])
                self.size_xy = np.array([it[6] for it in content])
                self.flick = np.array([it[7] for it in content])
                self.interpolate = np.array([it[8] for it in content], dtype=bool)
                self.mask = np.array([it[9] is not None for it in content], dtype=bool)
                self.sf = np.array([it[10] for it in content])
                self.tf = np.array([it[11] for it in content])
                self.opto = np.array([it[12] for it in content], dtype=int)
                self.pattern = np.array([it[13] for it in content])

                remove_code.append(code)

            elif self.log_info[code] == 'PhotoIndicator':
                self.photo_time = np.array([it[0] for it in content])
                self.photo_state = np.array([it[1] for it in content], dtype=bool)
                self.photo_size = np.array([it[2] for it in content])
                self.photo_pos = np.array([it[3] for it in content])
                self.photo_units = np.array([it[4] for it in content])
                self.photo_mode = np.array([it[5] for it in content], dtype=int)
                self.photo_frames = np.array([it[6] for it in content], dtype=int)
                self.photo_enable = np.array([it[7] for it in content], dtype=bool)

                remove_code.append(code)

        for code in remove_code:
            del log_data[code]

        return log_data

    def get_visual_stim_dataframe(self) -> pl.DataFrame:
        """
        Visual presentation dataframe::

            ┌────────────┬──────────┬──────────┬─────┬───┬──────┬─────┬──────┬─────────┐
            │ time       ┆ duration ┆ contrast ┆ ori ┆ … ┆ sf   ┆ tf  ┆ opto ┆ pattern │
            │ ---        ┆ ---      ┆ ---      ┆ --- ┆   ┆ ---  ┆ --- ┆ ---  ┆ ---     │
            │ f64        ┆ i64      ┆ i64      ┆ i64 ┆   ┆ f64  ┆ i64 ┆ i64  ┆ str     │
            ╞════════════╪══════════╪══════════╪═════╪═══╪══════╪═════╪══════╪═════════╡
            │ 18.990026  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
            │ 21.000029  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
            │ …          ┆ …        ┆ …        ┆ …   ┆ … ┆ …    ┆ …   ┆ …    ┆ …       │
            │ 619.054972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
            │ 619.084972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
            └────────────┴──────────┴──────────┴─────┴───┴──────┴─────┴──────┴─────────┘

        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(self.time).alias('time'),
            pl.Series(self.duration).alias('duration'),
            pl.Series(self.contrast).alias('contrast'),
            pl.Series(self.ori).alias('ori'),
            pl.Series(self.phase).alias('phase'),
            pl.Series(self.pos_xy).alias('pos'),
            pl.Series(self.size_xy).alias('size'),
            pl.Series(self.flick).alias('flick'),
            pl.Series(self.interpolate).alias('interpolate'),
            pl.Series(self.mask).alias('mask'),
            pl.Series(self.sf).alias('sf'),
            pl.Series(self.tf).alias('tf'),
            pl.Series(self.opto).alias('opto'),
            pl.Series(self.pattern).alias('pattern')
        )
        return df

    def get_photo_indicator_dataframe(self) -> pl.DataFrame:
        """
        PhotoDiode dataframe::

            ┌────────────┬───────┬──────┬────────────┬───────┬──────┬────────┬────────┐
            │ time       ┆ state ┆ size ┆ pos        ┆ units ┆ mode ┆ frames ┆ enable │
            │ ---        ┆ ---   ┆ ---  ┆ ---        ┆ ---   ┆ ---  ┆ ---    ┆ ---    │
            │ f64        ┆ bool  ┆ i64  ┆ list[i64]  ┆ str   ┆ i64  ┆ i64    ┆ bool   │
            ╞════════════╪═══════╪══════╪════════════╪═══════╪══════╪════════╪════════╡
            │ 18.990026  ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ 21.000029  ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ …          ┆ …     ┆ …    ┆ …          ┆ …     ┆ …    ┆ …      ┆ …      │
            │ 607.094955 ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            │ 609.104958 ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
            └────────────┴───────┴──────┴────────────┴───────┴──────┴────────┴────────┘

        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(self.photo_time).alias('time'),
            pl.Series(self.photo_state).alias('state'),
            pl.Series(self.photo_size).alias('size'),
            pl.Series(self.photo_pos).alias('pos'),
            pl.Series(self.photo_units).alias('units'),
            pl.Series(self.photo_mode).alias('mode'),
            pl.Series(self.photo_frames).alias('frames'),
            pl.Series(self.photo_enable).alias('enable')
        )
        return df

    def get_state_machine_dataframe(self) -> pl.DataFrame:
        """
        State Machine dataframe::

            ┌────────────┬───────────────────────────┬───────────────────────────┐
            │ time       ┆ state                     ┆ prev_state                │
            │ ---        ┆ ---                       ┆ ---                       │
            │ f64        ┆ str                       ┆ str                       │
            ╞════════════╪═══════════════════════════╪═══════════════════════════╡
            │ 18.990026  ┆ ('States.SHOW_BLANK', 1)  ┆ ('States.STIM_SELECT', 0) │
            │ 20.990029  ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
            │ …          ┆ …                         ┆ …                         │
            │ 609.094958 ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
            │ 619.094972 ┆ ('States.STIM_SELECT', 0) ┆ ('States.SHOW_STIM', 2)   │
            └────────────┴───────────────────────────┴───────────────────────────┘

        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(it[0] for it in self.log_data[2]).alias('time'),
            pl.Series(it[1] for it in self.log_data[2]).alias('state'),
            pl.Series(it[2] for it in self.log_data[2]).alias('prev_state'),
        )
        return df

    def get_log_dict_dataframe(self) -> pl.DataFrame:
        """
        Log Dict DataFrame::

            ┌────────────┬──────────┬──────────┬──────────────┬────────────┐
            │ time       ┆ block_nr ┆ trial_nr ┆ condition_nr ┆ trial_type │
            │ ---        ┆ ---      ┆ ---      ┆ ---          ┆ ---        │
            │ f64        ┆ i64      ┆ i64      ┆ i64          ┆ i64        │
            ╞════════════╪══════════╪══════════╪══════════════╪════════════╡
            │ 18.990026  ┆ 0        ┆ 0        ┆ 0            ┆ 1          │
            │ 30.990043  ┆ 1        ┆ 0        ┆ 0            ┆ 1          │
            │ …          ┆ …        ┆ …        ┆ …            ┆ …          │
            │ 595.083939 ┆ 48       ┆ 0        ┆ 0            ┆ 1          │
            │ 607.094955 ┆ 49       ┆ 0        ┆ 0            ┆ 1          │
            └────────────┴──────────┴──────────┴──────────────┴────────────┘

        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(it[0] for it in self.log_data[3]).alias('time'),
            pl.Series(it[1] for it in self.log_data[3]).alias('block_nr'),
            pl.Series(it[2] for it in self.log_data[3]).alias('trial_nr'),
            pl.Series(it[3] for it in self.log_data[3]).alias('condition_nr'),
            pl.Series(it[4] for it in self.log_data[3]).alias('trial_type'),
        )

        return df

    def get_column(self,
                   category: int | str,
                   field: str,
                   dtype=float,
                   mapper: Callable[[Any], Any] | None = None) -> np.ndarray:

        if category in ('Gratings', 'FunctionBased', 'PhotoIndicator'):
            raise RuntimeError(f'preprocessed log : {category}:{field}')

        if isinstance(category, str):
            category = self._get_category_code(category)

        field = self._get_field_index(category, field)
        data = [it[field] for it in self.log_data[category]]

        if mapper is not None:
            # i.e., operator.itemgetter(1) for tuple statemachine ('States.SHOW_STIM', 2)...
            data = list(map(mapper, data))

        return np.array(data, dtype=dtype)

    def _get_category_code(self, category: str) -> int:
        for code, name in self.log_info.items():
            if name == category:
                return code
        raise KeyError('')

    def _get_field_index(self, category: int, field: str) -> int:
        header = self.log_header[category]
        if field == 'time':
            return 0
        return header.index(field) + 1

    @property
    def pos_x(self) -> np.ndarray:
        """object center position X. Array[float, P]"""
        return self.pos_xy[:, 0]

    @property
    def pos_y(self) -> np.ndarray:
        """object center position Y. Array[float, P]"""
        return self.pos_xy[:, 1]

    @property
    def size_x(self) -> np.ndarray:
        """object size width. Array[int, P]"""
        return self.size_xy[:, 0]

    @property
    def size_y(self) -> np.ndarray:
        """object size height. Array[int, P]"""
        return self.size_xy[:, 1]

    @property
    def exp_start_time(self) -> float:
        return self.riglog_data.dat[0, 2] / 1000

    @property
    def exp_end_time(self) -> float:
        return self.riglog_data.dat[-1, 2] / 1000

    @property
    def stim_start_time(self) -> float:
        return float(self.photo_time[1] + self.time_offset)

    @property
    def stim_end_time(self) -> float:
        return float(self.time[-1] + self.time_offset)

    @property
    def stimulus_segment(self) -> np.ndarray:
        # ret = self.riglog_data.screen_event[:, 0].reshape(-1, 2)  # directly use riglog time
        # stimlog bug starting from `diode off`, and ending from `diode True`
        # thus remove first point and add last point manually
        _p_time = np.concatenate((self.photo_time[1:], np.array([self.time[-1]])))
        ret = _p_time.reshape(-1, 2) + self.time_offset
        return ret

    def session_trials(self) -> dict[Session, SessionInfo]:
        return {
            prot.name: prot
            for prot in get_protocol_sessions(self)
        }

    @property
    def time_offset(self) -> float:
        if self._cache_time_offset is None:
            # noinspection PyTypeChecker
            self._cache_time_offset = _time_offset(self.riglog_data, self, self.diode_offset)[0]
        return self._cache_time_offset

    def get_stim_pattern(self) -> GratingPattern:
        prot = StimpyProtocol.load(self.stimlog_file.with_suffix('.prot'))
        log_nr = self.get_column('LogDict', 'condition_nr').astype(int)

        t = self.stimulus_segment
        ori = np.array([prot['ori'][n] for n in log_nr])
        sf = np.array([prot['sf'][n] for n in log_nr])
        tf = np.array([prot['tf'][n] for n in log_nr])
        contrast = np.array([prot['c'][n] for n in log_nr])
        dur = np.array([prot['dur'][n] for n in log_nr])

        return GratingPattern(t, ori, sf, tf, contrast, dur)

    @property
    def profile_dataframe(self) -> pl.DataFrame:
        raise NotImplementedError('')


# ========== #

@deprecated_func(remarks='generalized, to sequential offset method')
def _time_offset(rig: RiglogData,
                 stm: StimlogGit,
                 diode_offset=True) -> tuple[float, float]:
    """
    time offset used in the `new stimpy`
    offset time from screen_time .riglog (diode) to .stimlog
    ** normally stimlog time value are larger than riglog

    :param rig:
    :param stm:
    :param diode_offset: whether correct diode signal
    :return: tuple of offset average and std value
    """
    if not isinstance(stm, StimlogGit):
        raise TypeError('')

    screen_time = rig.screen_event.time

    if diode_offset:
        try:
            # new stimpy might be a negative value (stimlog time value larger than riglog)
            # stimlog bug starting from `diode off`, and ending from `diode True`
            # thus remove first point and add last point manually
            _p_time = np.concatenate((stm.photo_time[1:], np.array([stm.time[-1]])))
            offset_t = screen_time[::2] - _p_time[::2]

        except ValueError as e:
            fprint(f'number of diode pulse and stimulus mismatch from {e}', vtype='error')
            fprint(f'use the first pulse diff for alignment', vtype='error')
            offset_t = screen_time[0] - stm.photo_time[1]
    else:
        raise NotImplementedError('')

    offset_t_avg = float(np.mean(offset_t))
    offset_t_std = float(np.std(offset_t))

    fprint(f'time offset between stimlog and riglog: {round(offset_t_avg, 3)}')
    fprint(f'offset_std: {round(offset_t_std, 3)}')

    return offset_t_avg, offset_t_std


def load_stimlog(file: PathLike, string_key: bool = True) -> dict[str | int, pl.DataFrame]:
    """
    Load directly the stimlog file (without riglog time offset), and parse data as polars dataframes

    .. code-block:: python

        file = ...

        log = load_stimlog(file, string_key=True)  # get dataframe using keyname
        print(log['PhotoIndicator'])

        log = load_stimlog(file, string_key=False)  # get dataframe using code int
        print(log[1])

    :param file: file path for the .stimlog
    :param string_key: show key as str type, otherwise, int type
    :return: Code:DataFrame dictionary
    """

    log_info = {}
    log_header = {}
    log_data = {}

    with Path(file).open() as f:
        for line, content in enumerate(f):
            content = content.strip()

            #
            m = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
            if m:
                info_name = m.group(1)
                info_value = m.group(2)
                try:
                    code = int(info_name)
                except ValueError:
                    continue
                else:
                    name, header = info_value.split(' ', maxsplit=1)
                    header = eval(header)
                    log_info[code] = name
                    log_header[code] = header

            elif content.startswith('#'):
                print(f'ignore header line at {line + 1} : {content}')
                continue

            else:  # data
                code, time, message = content.split(' ', maxsplit=2)
                code = int(code)
                time = float(time)

                if log_info[code] == 'StateMachine':
                    value = eval(message.replace('<', '("').replace(':', '",').replace('>', ')'))
                    value = list(map(str, value))
                    if len(log_header[code]) != len(value):
                        print(f'log category {code} size mismatched at line {line} : {message}')
                    else:
                        log_data.setdefault(code, []).append((time, *value))

                else:
                    value = eval(message)
                    if len(log_header[code]) != len(value):
                        print(f'log category {code} size mismatched at line {line} : {message}')
                    else:
                        log_data.setdefault(code, []).append((time, *value))

        return {
            (log_info[code] if string_key else code):
                pl.DataFrame(log_data[code], schema=['time'] + log_header[code], strict=False)
            for code in log_data
        }
